@@ -443,7 +443,7 @@ static Xen xen_rb_rep(Xen ig)
   Xen val;
   char *str;
   size_t size = 512;
-  char **buffer = NULL;
+  char **buffer;
   buffer = (char **)calloc(1, sizeof(char *));
   buffer[0] = (char *)calloc(size, sizeof(char));
   fprintf(stdout, "%s", rb_prompt);
@@ -502,13 +502,23 @@ Xen xen_rb_eval_string_with_error(const char *str)
 }
 
 
-Xen xen_rb_load_file_with_error(Xen file)
+void xen_rb_load_file_with_error(const char *file)
 {
-  int status = 0;
-  rb_load_protect(file, 0, &status);
-  if (status != 0)
-    return(xen_rb_obj_as_string(rb_gv_get("$!")));
-  return(Xen_true);
+  int status = 0, i;
+  Xen err, info;
+  rb_load_protect(C_string_to_Xen_string(file), 0, &status);
+  if (status == 0)
+    return;
+  fprintf(stderr, "Can't load %s", file);
+  err = rb_gv_get("$!");
+  if (err != Qnil)
+    fprintf(stderr, ": %s", Xen_object_to_C_string(err));
+  fprintf(stderr, "\n");
+  info = rb_gv_get("$@");
+  if (info == Qnil)
+    return;
+  for (i = 0; i < Xen_vector_length(info); i++)
+    fprintf(stderr, "%s\n", Xen_string_to_C_string(Xen_vector_ref(info, i)));
 }
 
 
@@ -1320,7 +1330,7 @@ void xen_repl(int argc, char **argv)
 {
   int size = 512;
   bool expr_ok = true;
-  char *buffer = NULL;
+  char *buffer;
   buffer = (char *)calloc(size, sizeof(char));
 
   while (true)
@@ -1370,13 +1380,14 @@ void xen_repl(int argc, char **argv)
 	    }
 	}
     }
+  /* unreachable */
   free(buffer);
 }
 
 
 void xen_gc_mark(Xen val)
 {
-  s7_mark_c_object(val);
+  s7_mark(val);
 }
 
 
@@ -1552,6 +1563,7 @@ static Xen g_localtime(Xen tm)
 {
   #define H_localtime "(localtime tm) breaks up tm into something suitable for strftime"
   time_t rtime;
+  Xen_check_type(Xen_is_integer(tm), tm, 1, "localtime", "an integer");
   rtime = (time_t)Xen_ulong_to_C_ulong(tm);
   return(Xen_wrap_C_pointer(localtime((time_t *)(&rtime))));
 }
@@ -1565,53 +1577,10 @@ static Xen g_current_time(void)
   return(C_ulong_to_Xen_ulong(curtime));
 }
 
-#if (!DISABLE_DEPRECATED)
-static Xen g_tmpnam(void)
-{
-  #define H_tmpnam "(tmpnam) returns a new (hopefully unused) temporary file name"
-  #define BUFFER_SIZE 512
-  static int file_ctr = 0;
-  char *str, *tmpdir = NULL;
-  Xen result;
-
-  str = (char *)calloc(BUFFER_SIZE, sizeof(char));
-  tmpdir = xen_strdup(getenv("TMPDIR"));
-
-#ifdef P_tmpdir
-  if (!tmpdir) 
-    tmpdir = xen_strdup(P_tmpdir); /* /usr/include/stdio.h */
-  if (tmpdir)
-    {
-      int len;
-      len = strlen(tmpdir);
-      if (len > 0)
-	{
-	  if (tmpdir[len - 1] == '/') 
-	    tmpdir[len - 1] = 0;
-	}
-      else
-	{
-	  free(tmpdir);
-	  tmpdir = xen_strdup(".");
-	}
-    }
-#else
-  if (!tmpdir) tmpdir = xen_strdup("/tmp");
-#endif
-
-  if (tmpdir) /* try to make C happy... */
-    snprintf(str, BUFFER_SIZE, "%s/xen_%d_%d", tmpdir, (int)getpid(), file_ctr++);
-  else snprintf(str, BUFFER_SIZE, "/xen_%d_%d", (int)getpid(), file_ctr++);
-  if (tmpdir) free(tmpdir);
-  result = C_string_to_Xen_string(str);
-  free(str);
-  return(result);
-}
-#endif
-
 
 static Xen g_ftell(Xen fd)
 {
+  Xen_check_type(Xen_is_integer(fd), fd, 1, "ftell", "an integer");
   return(C_int_to_Xen_integer(lseek(Xen_integer_to_C_int(fd), 0, SEEK_CUR)));
 }
 
@@ -1648,9 +1617,6 @@ Xen_wrap_no_args(g_current_time_w, g_current_time)
 Xen_wrap_1_arg(g_ftell_w, g_ftell)
 Xen_wrap_no_args(g_gc_off_w, g_gc_off)
 Xen_wrap_no_args(g_gc_on_w, g_gc_on)
-#if (!DISABLE_DEPRECATED)
-Xen_wrap_no_args(g_tmpnam_w, g_tmpnam)
-#endif
 
 #if ENABLE_WEBSERVER
   #if USE_MOTIF
@@ -1735,12 +1701,6 @@ s7_scheme *s7_xen_initialize(s7_scheme *sc)
 		                     ((eq? func (car l)) (loop (cdr l) result))\n\
 		                     (else (loop (cdr l) (cons (car l) result)))))))");
 
-#if (!DISABLE_DEPRECATED)
-  Xen_define_typed_procedure("tmpnam",       g_tmpnam_w,        0, 0, 0, H_tmpnam,        s7_make_signature(s7, 1, s));
-  Xen_eval_C_string("(define load-from-path load)");
-  Xen_eval_C_string("(define (1+ x) \"add 1 to arg\" (+ x 1))");
-  Xen_eval_C_string("(define (1- x) \"subtract 1 from arg\" (- x 1))");
-#endif
   Xen_eval_C_string("(define-macro (while whether . body) `(do () ((not ,whether)) ,@body))");
   Xen_eval_C_string("(define (identity x) \"return arg\" x)");
 
