@@ -1,8 +1,10 @@
 class Sbcl < Formula
   desc "Steel Bank Common Lisp system"
   homepage "http://www.sbcl.org/"
-  url "https://downloads.sourceforge.net/project/sbcl/sbcl/1.5.7/sbcl-1.5.7-source.tar.bz2"
-  sha256 "54742fb5e2f3f350fbafd72bc73662fca21382b5553ed6a146098950d2409485"
+  url "https://downloads.sourceforge.net/project/sbcl/sbcl/1.4.11/sbcl-1.4.11-source.tar.bz2"
+  sha256 "f4b82e95ed1b1d973901442fb9e609c1446bc50a4e554ca8b94b8e8c281c4eb5"
+
+  option "with-internal-xref", "Include XREF information for SBCL internals (increases core size by 5-6MB)"
 
   # Current binary versions are listed at https://sbcl.sourceforge.io/platform-table.html
   resource "bootstrap64" do
@@ -38,6 +40,7 @@ class Sbcl < Formula
       "--with-sb-ldb",
       "--with-sb-thread",
     ]
+    args << "--with-sb-xref-internal" if build.with? "internal-xref"
 
     ENV["SBCL_MACOSX_VERSION_MIN"] = MacOS.version
     system "./make.sh", *args
