@@ -9,7 +9,7 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB-VM")
+(in-package "SB!VM")
 
 ;;;; allocation helpers
 
@@ -62,34 +62,34 @@
 (defun allocation-inline (alloc-tn size)
   (let* ((ok (gen-label))
          (done (gen-label))
-         #+(and sb-thread win32)
+         #!+(and sb-thread win32)
          (scratch-tns (loop for my-tn in `(,eax-tn ,ebx-tn ,edx-tn ,ecx-tn)
                             when (and (not (location= alloc-tn my-tn))
                                       (or (not (tn-p size))
                                           (not (location= size my-tn))))
                             collect my-tn))
-         (tls-prefix #+sb-thread :fs)
-         #+(and sb-thread win32) (scratch-tn (pop scratch-tns))
-         #+(and sb-thread win32) (swap-tn (pop scratch-tns))
+         (tls-prefix #!+sb-thread :fs)
+         #!+(and sb-thread win32) (scratch-tn (pop scratch-tns))
+         #!+(and sb-thread win32) (swap-tn (pop scratch-tns))
          (free-pointer
            ;; thread->alloc_region.free_pointer
            (make-ea :dword
-                    :base (or #+(and sb-thread win32)
+                    :base (or #!+(and sb-thread win32)
                               scratch-tn)
                     :disp
-                    #+sb-thread (* n-word-bytes thread-alloc-region-slot)
-                    #-sb-thread (make-fixup "gc_alloc_region" :foreign)))
+                    #!+sb-thread (* n-word-bytes thread-alloc-region-slot)
+                    #!-sb-thread (make-fixup "gc_alloc_region" :foreign)))
          (end-addr
             ;; thread->alloc_region.end_addr
            (make-ea :dword
-                    :base (or #+(and sb-thread win32)
+                    :base (or #!+(and sb-thread win32)
                               scratch-tn)
                     :disp
-                    #+sb-thread (* n-word-bytes (1+ thread-alloc-region-slot))
-                    #-sb-thread (make-fixup "gc_alloc_region" :foreign 4))))
+                    #!+sb-thread (* n-word-bytes (1+ thread-alloc-region-slot))
+                    #!-sb-thread (make-fixup "gc_alloc_region" :foreign 4))))
     (unless (and (tn-p size) (location= alloc-tn size))
       (inst mov alloc-tn size))
-    #+(and sb-thread win32)
+    #!+(and sb-thread win32)
     (progn
       (inst push scratch-tn)
       (inst push swap-tn)
@@ -113,12 +113,12 @@
     ;; Swap ALLOC-TN and FREE-POINTER
     (cond ((and (tn-p size) (location= alloc-tn size))
            ;; XCHG is extremely slow, use the xor swap trick
-           #-(and sb-thread win32)
+           #!-(and sb-thread win32)
            (progn
              (inst xor alloc-tn free-pointer tls-prefix)
              (inst xor free-pointer alloc-tn tls-prefix)
              (inst xor alloc-tn free-pointer tls-prefix))
-           #+(and sb-thread win32)
+           #!+(and sb-thread win32)
            (progn
              (inst mov swap-tn free-pointer tls-prefix)
              (inst mov free-pointer alloc-tn tls-prefix)
@@ -128,7 +128,7 @@
            (inst mov free-pointer alloc-tn tls-prefix)
            (inst sub alloc-tn size)))
     (emit-label done)
-    #+(and sb-thread win32)
+    #!+(and sb-thread win32)
     (progn
       (inst pop swap-tn)
       (inst pop scratch-tn))
@@ -151,7 +151,7 @@
      (stack-allocation alloc-tn size lowtag))
     ;; Inline allocation can't work if (and (not sb-thread) sb-dynamic-core)
     ;; because boxed_region points to the linkage table, not the alloc region.
-    #+(or sb-thread (not sb-dynamic-core))
+    #!+(or sb-thread (not sb-dynamic-core))
     ((or (null node) (policy node (>= speed space)))
      (allocation-inline alloc-tn size))
     (t
@@ -163,11 +163,11 @@
 ;;; Allocate an other-pointer object of fixed SIZE with a single word
 ;;; header having the specified WIDETAG value. The result is placed in
 ;;; RESULT-TN.
-(defun alloc-other (result-tn widetag size node &optional stack-allocate-p)
+(defun fixed-alloc (result-tn widetag size node &optional stack-allocate-p)
   (pseudo-atomic (:elide-if stack-allocate-p)
       (allocation result-tn (pad-data-block size) node stack-allocate-p
                   other-pointer-lowtag)
-      (storew (compute-object-header size widetag)
+      (storew (logior (ash (1- size) n-widetag-bits) widetag)
               result-tn 0 other-pointer-lowtag)))
 
 ;;;; CONS, LIST and LIST*
@@ -256,8 +256,8 @@
        (inst lea result (make-ea :byte :base result :disp other-pointer-lowtag))
        (sc-case type
          (immediate
-          (aver (typep (tn-value type) '(unsigned-byte 9)))
-          (store-widetag (tn-value type) result 0 other-pointer-lowtag))
+          (aver (typep (tn-value type) '(unsigned-byte 8)))
+          (storeb (tn-value type) result 0 other-pointer-lowtag))
          (t
           (storew type result 0 other-pointer-lowtag)))
        (sc-case length
@@ -265,7 +265,7 @@
           (let ((fixnum-length (fixnumize (tn-value length))))
             (typecase fixnum-length
               ((unsigned-byte 8)
-               (store-widetag fixnum-length result
+               (storeb fixnum-length result
                        vector-length-slot other-pointer-lowtag))
               (t
                (storew fixnum-length result
@@ -302,7 +302,7 @@
     (sc-case type
       (immediate
        (aver (typep (tn-value type) '(unsigned-byte 8)))
-       (store-widetag (tn-value type) result 0 other-pointer-lowtag))
+       (storeb (tn-value type) result 0 other-pointer-lowtag))
       (t
        (storew type result 0 other-pointer-lowtag)))
     (storew length result vector-length-slot other-pointer-lowtag)
@@ -318,7 +318,7 @@
   (:results (result :scs (descriptor-reg) :from :argument))
   (:node-var node)
   (:generator 37
-    (alloc-other result fdefn-widetag fdefn-size node)
+    (fixed-alloc result fdefn-widetag fdefn-size node)
     (storew name result fdefn-name-slot other-pointer-lowtag)
     (storew nil-value result fdefn-fun-slot other-pointer-lowtag)
     (storew (make-fixup 'undefined-tramp :assembly-routine)
@@ -338,11 +338,14 @@
                    stack-allocate-p
                    fun-pointer-lowtag)
        (storew (logior (ash (1- size) n-widetag-bits) closure-widetag)
-               result 0 fun-pointer-lowtag)))
-   ;; Done with pseudo-atomic
-   (inst lea temp (make-ea-for-object-slot function simple-fun-insts-offset
-                                           fun-pointer-lowtag))
-   (storew temp result closure-fun-slot fun-pointer-lowtag)))
+               result 0 fun-pointer-lowtag))
+    ;; These two instructions are within the scope of PSEUDO-ATOMIC.
+    ;; This is due to scav_closure() assuming that it can always subtract
+    ;; FUN_RAW_ADDR_OFFSET from closure->fun to obtain a Lisp object,
+    ;; without any precheck for whether that word is currently 0.
+    (inst lea temp (make-ea-for-object-slot function simple-fun-code-offset
+                                            fun-pointer-lowtag))
+    (storew temp result closure-fun-slot fun-pointer-lowtag))))
 
 ;;; The compiler likes to be able to directly make value cells.
 (define-vop (make-value-cell)
@@ -351,7 +354,7 @@
   (:info stack-allocate-p)
   (:node-var node)
   (:generator 10
-    (alloc-other result value-cell-widetag value-cell-size node stack-allocate-p)
+    (fixed-alloc result value-cell-widetag value-cell-size node stack-allocate-p)
     (storew value result value-cell-value-slot other-pointer-lowtag)))
 
 ;;;; automatic allocators for primitive objects
